from dune.perftool.generation import (include_file,
                                      preamble,
                                      )
from dune.perftool.options import get_option
from dune.perftool.pdelab.driver import (form_name_suffix,
                                         get_formdata,
                                         is_linear,
                                         name_initree,
                                         )
from dune.perftool.pdelab.driver.gridfunctionspace import name_trial_gfs
from dune.perftool.pdelab.driver.gridoperator import (name_gridoperator,
                                                      type_gridoperator,
                                                      )
from dune.perftool.pdelab.driver.interpolate import interpolate_dirichlet_data


@preamble
def dune_solve():
    # Test if form is linear in ansatzfunction
    linear = is_linear()

    # Test wether we want to do matrix free operator evaluation
    matrix_free = get_option('matrix_free')

    # Get right solve command
    if linear and matrix_free:
        formdata = get_formdata()
        go = name_gridoperator(formdata)
        x = name_vector(formdata)
        include_file("dune/perftool/matrixfree.hh", filetag="driver")
        solve = "solveMatrixFree({},{});".format(go, x)
    elif linear and not matrix_free:
        slp = name_stationarylinearproblemsolver()
        solve = "{}.apply();".format(slp)
    elif not linear and matrix_free:
        # TODO copy of linear case and obviously broken, used to generate something ;)
        formdata = get_formdata()
        go = name_gridoperator(formdata)
        x = name_vector(formdata)
        include_file("dune/perftool/matrixfree.hh", filetag="driver")
        solve = "solveNonlinearMatrixFree({},{});".format(go, x)
    elif not linear and not matrix_free:
        go_type = type_gridoperator(get_formdata())
        go = name_gridoperator(get_formdata())
        snp = name_stationarynonlinearproblemsolver(go_type, go)
        solve = "{}.apply();".format(snp)

    print_residual()
    print_matrix()

    if get_option('instrumentation_level') >= 2:
        from dune.perftool.pdelab.driver.timings import setup_timer, name_timing_stream
        setup_timer()
        from dune.perftool.generation import post_include
        post_include("HP_DECLARE_TIMER(solve);", filetag="driver")

        # Print times after solving
        from dune.perftool.generation import get_global_context_value
        formdatas = get_global_context_value("formdatas")
        print_times = []
        for formdata in formdatas:
            from dune.perftool.pdelab.driver.gridoperator import name_localoperator
            lop_name = name_localoperator(formdata)
            timestream = name_timing_stream()
            print_times.append("{}.dump_timers({}, {}, true);".format(lop_name, timestream, name_timing_identifier()))

        solve = ["HP_TIMER_START(solve);",
                 "{}".format(solve),
                 "HP_TIMER_STOP(solve);",
                 "DUMP_TIMER({}, solve, {}, true);".format(get_option("instrumentation_level"), timestream),
                 ]
        if get_option('instrumentation_level') >= 3:
            solve.extend(print_times)

    return solve


def name_vector(formdata):
    name = form_name_suffix("x", formdata)
    define_vector(name, formdata)
    interpolate_dirichlet_data(name)
    return name


@preamble
def typedef_vector(name, formdata):
    go_type = type_gridoperator(formdata)
    return "using {} = {}::Traits::Domain;".format(name, go_type)


def type_vector(formdata):
    name = form_name_suffix("V", formdata).upper()
    typedef_vector(name, formdata)
    return name


@preamble
def define_vector(name, formdata):
    vtype = type_vector(formdata)
    gfs = name_trial_gfs()
    return ["{} {}({});".format(vtype, name, gfs), "{} = 0.0;".format(name)]


@preamble
def typedef_linearsolver(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "using {} = Dune::PDELab::ISTLBackend_SEQ_SuperLU;".format(name)


def type_linearsolver():
    name = "LinearSolver"
    typedef_linearsolver(name)
    return name


@preamble
def define_linearsolver(name):
    lstype = type_linearsolver()
    return "{} {}(false);".format(lstype, name)


def name_linearsolver():
    name = "ls"
    define_linearsolver(name)
    return name


@preamble
def define_reduction(name):
    ini = name_initree()
    return "double {} = {}.get<double>(\"reduction\", 1e-12);".format(name, ini)


def name_reduction():
    name = "reduction"
    define_reduction(name)
    return name


@preamble
def typedef_stationarylinearproblemsolver(name):
    include_file("dune/pdelab/stationary/linearproblem.hh", filetag="driver")
    gotype = type_gridoperator(get_formdata())
    lstype = type_linearsolver()
    xtype = type_vector(get_formdata())
    return "using {} = Dune::PDELab::StationaryLinearProblemSolver<{}, {}, {}>;".format(name, gotype, lstype, xtype)


def type_stationarylinearproblemsolver():
    typedef_stationarylinearproblemsolver("SLP")
    return "SLP"


@preamble
def define_stationarylinearproblemsolver(name):
    slptype = type_stationarylinearproblemsolver()
    formdata = get_formdata()
    go = name_gridoperator(formdata)
    ls = name_linearsolver()
    x = name_vector(formdata)
    red = name_reduction()
    return "{} {}({}, {}, {}, {});".format(slptype, name, go, ls, x, red)


def name_stationarylinearproblemsolver():
    define_stationarylinearproblemsolver("slp")
    return "slp"


@preamble
def typedef_stationarynonlinearproblemsolver(name, go_type):
    include_file("dune/pdelab/newton/newton.hh", filetag="driver")
    ls_type = type_linearsolver()
    x_type = type_vector(get_formdata())
    return "using {} = Dune::PDELab::Newton<{}, {}, {}>;".format(name, go_type, ls_type, x_type)


def type_stationarynonlinearproblemssolver(go_type):
    name = "SNP"
    typedef_stationarynonlinearproblemsolver(name, go_type)
    return name


@preamble
def define_stationarynonlinearproblemsolver(name, go_type, go):
    snptype = type_stationarynonlinearproblemssolver(go_type)
    x = name_vector(get_formdata())
    ls = name_linearsolver()
    return "{} {}({}, {}, {});".format(snptype, name, go, x, ls)


def name_stationarynonlinearproblemsolver(go_type, go):
    name = "snp"
    define_stationarynonlinearproblemsolver(name, go_type, go)
    return name


@preamble
def print_residual():
    ini = name_initree()
    formdata = get_formdata()
    n_go = name_gridoperator(formdata)
    v = name_vector(formdata)
    t_v = type_vector(formdata)
    include_file("random", system=True, filetag="driver")

    return ["if ({}.get<bool>(\"printresidual\", false)) {{".format(ini),
            "  using Dune::PDELab::Backend::native;",
            "  {} r({});".format(t_v, v),
            "  // Setup random input",
            "  std::size_t seed = 0;",
            "  auto rng = std::mt19937_64(seed);",
            "  auto dist = std::uniform_real_distribution<>(-1., 1.);",
            "  for (auto& v : {})".format(v),
            "    v = dist(rng);",
            "  r=0.0;",
            "  {}.residual({}, r);".format(n_go, v),
            "  Dune::printvector(std::cout, native(r), \"residual vector\", \"row\");",
            "}"]


@preamble
def print_matrix():
    formdata = get_formdata()
    ini = name_initree()
    t_go = type_gridoperator(formdata)
    n_go = name_gridoperator(formdata)
    v = name_vector(formdata)
    t_v = type_vector(formdata)

    return ["if ({}.get<bool>(\"printmatrix\", false)) {{".format(ini),
            "  // Setup random input",
            "  std::size_t seed = 0;",
            "  auto rng = std::mt19937_64(seed);",
            "  auto dist = std::uniform_real_distribution<>(1., 10.);",
            "  for (auto& v : {})".format(v),
            "    v = dist(rng);",
            "  using M = typename {}::Traits::Jacobian;".format(t_go),
            "  M m({});".format(n_go),
            "  {}.jacobian({},m);".format(n_go, v),
            "  using Dune::PDELab::Backend::native;",
            "  Dune::printmatrix(std::cout, native(m),\"global stiffness matrix\",\"row\",9,1);",
            "}"]
